import { LicenseInfo } from '../parser/types';
import { licenses, Licenses, License, exceptions, Exceptions, Exception } from './data';
export { licenses, Licenses, License, exceptions, Exceptions, Exception };
type PartialExpandLicenseOptions = {
    expandScope?: boolean;
};
export declare const expandLicenses: (identifiers: string[], providedOptions?: PartialExpandLicenseOptions) => string[];
/**
 * Fix the given license identifier, if possible, or return it unchanged.
 *
 * @param identifier SPDX license identifier to "fix" if possible
 * @returns A corrected SPDX license identifier or the original, if no suitable "fix" was found
 */
export declare function correctLicenseId(identifier: string, upgradeGPLVariants: boolean): string;
/**
 * Fix the given exception identifier, if possible, or return it unchanged.
 *
 * @param identifier SPDX exception identifier to "fix" if possible
 * @returns A corrected SPDX exception identifier or the original, if no suitable "fix" was found
 */
export declare function correctExceptionId(identifier: string, associatedLicense?: string): string;
export declare function isKnownLicenseId(id: string): boolean;
export declare function isKnownExceptionId(id: string): boolean;
export declare function fixDashedLicenseInfo(node: LicenseInfo, upgradeGPLVariants: boolean): LicenseInfo;
export declare const variationsOf: (name: string, upgradeGPLVariants: boolean) => string[];
/**
 * Try to identify the license by comparing the given text with the official names of
 * known licenses in the SPDX data.
 *
 * @param text The text that could be a license name
 * @returns The {@link License}, or `undefined`
 */
export declare const findNameBasedMatch: (text: string, upgradeGPLVariants: boolean) => License | undefined;
