import { ParsedSpdxExpression, ConjunctionInfo, LicenseInfo, LicenseRef } from './types';
export { ParsedSpdxExpression, ConjunctionInfo, LicenseInfo, LicenseRef };
export type FullSpdxParseResult = {
    input: string;
    error?: string;
    ast?: any;
    expression?: ParsedSpdxExpression;
};
export type ParseOptions = {
    strictSyntax?: boolean;
    upgradeGPLVariants?: boolean;
};
/**
 * Parse an SPDX expression into a structured object representation.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {ParsedSpdxExpression} A structured object describing the given SPDX expression
 *          or throws an `Error` if parsing failed.
 */
export declare function parse(input: string, parseOptions?: ParseOptions): ParsedSpdxExpression;
/**
 * Parse an SPDX expression into a structured object representation along with additional
 * metadata such as the underlying AST tree used and any errors in case parsing failed.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {FullSpdxParseResult}
 */
export declare function parseSpdxExpressionWithDetails(input: string, strictSyntax?: boolean, upgradeGPLVariants?: boolean): FullSpdxParseResult;
